      program p1ddcn
c     This is a MAIN PROGRAM for the solution to the Dirichlet boundary
c     value problem for the Poisson equation
c                    -d(K du/dx)/dx=f
c                     u(a)=psia, u(b)=psib
c      for a 1-D case.
c=================================================================
c    Discretization - SCALAR functions in CELLS ( HC )
c                   - VECTOR functions in NODES ( HN ) 
c=================================================================
      implicit real*8 (a-h,o-z)
c============================================================================

c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps

c     Set of coordinates of grid points
c     call subroutine mkgrid

      call mkgrid


c     Set boundary conditions 
c     call subroutine bcon
      call bcon     

c     Computation of coefficient K in cells
c     call subroutine coeff
      call coeff

c     Computation of the right-hand side of the differential equation in nodes
c     call subroutine right 
      call right


c     computation of coefficients and the right-hand side of the
c     finite-difference equations
      call CFDS

c     Set of constants for the termination of the iteration process
      eps=1.d-12

c     solution for the finite-difference equations by the GAUSS-SIEDEL method
      call siedel 

c     Computing the error
      call error

c--------------------------------------------------------------
c      Test output
c--------------------------------------------------------------
       call output
      end

c**************************************************************

      subroutine CFDS
      implicit real*8 (a-h,o-z)
c     This is subroutine CFDS for the computation of coefficients
c     and the right-hand side of the finite-fifference scheme for 
c     the Dirichlet boundary problem for the Poisson equation 
c                    -d(K du/dx)/dx=f
c                     u(a)=psia, u(b)=psib
c      for 1-D case.
c=================================================================
c     Discretization - Scalar functions in Nodes ( HN ),
c                      Vector functions in Cells ( HC )
c=================================================================
      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     ZK - array (1 to M), value of coefficient K in nodes
      common/k/ zk(200)

c     F - array (1 to M-1), values for the right-hand side of the
c                           differential equation in cells
      common/f/ f(200)

c     psia, psib - constants in the Dirichlet boundary conditions in
c     points a and b respectively.
      common/psi/ psia, psib

c     =================================================================
c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 2 to M-1 ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)
c     =================================================================

c     VC, VN - arrays for the volume of cells and nodes
      dimension VC(200), VN(200)
c     =================================================================

c     Computing VC and VN

      do 1 i=1,M-1
      VC(i) = x(i+1)-x(i)
1     continue

      VN(1)=0.5*(x(2)-x(1))
      do 2 i=2,M-1
      VN(i)=0.5*(x(i+1)-x(i-1))
2     continue
      VN(M)=0.5*(x(M)-x(M-1))
c

c     computing the coefficients for the finite-difference scheme
c     and the right-hand side
c     ========================================================

      c(1)=zk(2)/vn(2)+zk(1)/vn(1)
      e(1)=-zk(2)/vn(2)
c VRODE NADO VC(1)
      phi(1)=f(1)*VC(2)+zk(1)*psia/vn(1)

      do 3 i=2,M-2
      w(i)=-zk(i)/vn(i)
      c(i)=zk(i)/vn(i)+zk(i+1)/vn(i+1)
      e(i)=-zk(i+1)/vn(i+1)
      phi(i)=f(i) * VC(i)
3     continue

      w(M-1)=-zk(M-1)/VN(M-1)
      c(M-1)=zk(M-1)/VN(M-1)+zk(M)/VN(M)
      phi(M-1)=f(M-1)*VC(M-1)+zk(M)*psib/VN(M)

      return
      end
c*************************************************************

      subroutine siedel
      implicit real*8 (a-h,o-z)
c
c     SOLUTION for the system of linear algebraic equations
c     by the GAUSS-SIEDEL iteration method
c
c     M - number of nodes in the grid
      common/M/ M

c     U -array (1 to M-1 ), solution for the finite-difference scheme.
      common/U/ u(200)

c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 1 to M-1 ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)

c     psia, psib - constants in the Dirichlet boundary conditions in
c     points a and b respectively.
      common/psi/ psia, psib

c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps
c============================================================================
c     SET an INITIAL value for the iteration solution
      do 4 i=1, M-1
      U(i)=0.d0
4     continue
c     beginning of the iteration process
5     continue

      USP1=(phi(1)-e(1)*U(2))/c(1)
      EPSMAX=dabs(USP1-U(1))
      U(1)=USP1

      do 6 i=2, M-2
      USP1=( phi(i)-w(i)*U(i-1)-e(i)*U(i+1))/c(i)
      EPSMAX=amax1(EPSMAX,dabs(USP1 - U(i)))
      U(i)=USP1
6     continue

      USP1=(phi(M-1)-w(M-1)*U(M-2))/c(M-1)
      EPSMAX=amax1(EPSMAX,dabs(USP1-U(M-1)))
      U(M-1)=USP1

c     checking the termination condition
      if(EPSMAX.GT.EPS) goto 5
      return
      end
c************************************************************

      subroutine mkgrid

      implicit real*8 (a-h,o-z)
c     This subroutine generates a grid in 1-D
c     That is, it generates array X (1 to M) of
c     the coordinates of the grid nodes

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     Set number of nodes
      M=17      

c     Set end points of the computational domain
      a=0.d0
      b=1.d0

c     Simple smooth grid as in the examples of the book

      do i=1,m
      xi=a+(b-a)*(i-1.d0)/(M-1)
      x(i)=xi**2
      end do
      return
      end
c************************************************************

      subroutine exact
      implicit real*8 (a-h,o-z)
c     Compute the exact soulution in the nodes
c     Results - array uex (1 to M)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     UEX - array (1 to M-1) exact solution in the nodes
      common/uex/ uex(200)
      
       pi=acos(-1.)

      do i=1,M-1
       uex(i)=sin(2*pi*0.5*(x(i)+x(i+1)))
      end do  
      return 
      end
c************************************************************

       subroutine output
       implicit real*8 (a-h,o-z)
c      Output subroutine 

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     u -array (1 to M-1 ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M-1) exact solution in the nodes
      common/uex/ uex(200)

c     ERR - array (1 to M-1), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c      Max-norm and L2-norm of the error between
c      the exact and approximate solutions 
       common/errors/ supn, zL2N

c       Open the file p1ddnc.output for
c       test output of program P1DDNC
       open(unit=1,file='p1ddcn.output')

       write(1,33)
33     format(15X, 'TEST OUTPUT FOR PROGRAM p1ddcn')
c     =========================================================
      write(1,*)

      write(1,8) M, supn, zL2N
8     format(2X,'Number of Nodes=',i5, 2X,
     1 'MAX-NORM = ', E12.5, 2X,'  L2-NORM = ', E12.5)
       write(1,*)

c     Print of the approximate and exact solutions of the test problem
       write(1,9) 
9      format(5X,'i',7X,' X',10X,
     1 'Appr.Sol.',4X,'Exact.Sol.',5X,'Error')
       do i=1,M-1
       write(1,7) i, x(i), u(i), uex(i), err(i)
       end do  
 7     format(1X, i5,2X, E12.5,2X, E12.5,2X, E12.5,2X, E12.5)
       return
       end
c**************************************************************

      subroutine error
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     VC - array ( 1 to M-1 ) volume of cells
c     COMMENTS - this volume is needed only for computing
c     the L2 Norm of error
      dimension VC(200)

c     u -array (1 to M-1 ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M-1) exact solution in the nodes
      common/uex/ uex(200)

c      Sup-norm and L2-norm of the error between
c      the exact and approximate solutions
       common/errors/ supn, zL2N

c     ERR - array (1 to M-1), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c      Computing the exact solution
       call exact

c     Computation of the volume of cells 

      do i=1,m-1
      vc(i)=x(i+1)-x(i)
      end do


c     Error
c     -----
      do i=1,M-1
      err(i)= uex(i)-u(i)
      end do     

c     sup - norm
c     -----------
      supn = 0.d0
      do i=1,M-1
      supn=amax1(supn,abs(err(i)))
      end do      

c     L  - norm
c      2
c     ----------

      zL2N = 0.d0
      do i=1, M-1
      zL2N = zL2N + (err(i))**2*VC(i)
      end do     
      zL2N = dsqrt(zL2N)
      return
      end
c****************************************************************
      
      subroutine bcon
      implicit real*8 (a-h,o-z)

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

      common/psi/ psia, psib
c     psia, psib - constants in the Dirichlet boundary conditions in 
c     points a and b respectively.
      psia=0.
      psib=0.
      return
      end
c*********************************************************************

      subroutine coeff
c     In this subroutine, the coefficients of the finite-difference
c     scheme in nodes must be computed
      implicit real*8 (a-h,o-z)


c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     ZK - array (1 to M), value of coefficient K in nodes
      common/k/ zk(200)


      do i=1,m
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
      zk(i)=fk(x(i))
      end do 
      return
      end
c****************************************************************

      real function fk*8(x)
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
     
      implicit real*8 (a-h,o-z)
      fk=x**4 + 1.
      return
      end
c****************************************************************

      subroutine right
c     This subroutine computes the values of the right-hand side
c     of the original differential equation in cells

      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)


c     F - array (1 to M-1), values for the right-hand side of the
c                           differential equation in cells
      common/f/ f(200)



      do i=1,m-1
       f(i)=fr(0.5*(x(i)+x(i+1)))
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x
      end do
      return 
      end
c*********************************************************************
      
      real function fr*8 (x)
      implicit real*8 (a-h,o-z)
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x

c     Number pi=3.14...
      pi=acos(-1.)
       fr=-2*pi*4.d0*x**3*cos(2*pi*x)
     1      +4*pi**2*(x**4+1)*sin(2*pi*x)
       return
       end
c**********************************************************************


