         program gid2de
         implicit real*8(a-h,o-z)
c        Main program for 2D explicit schemes
c        for Lagrangian gas dynamics.
c       
c        Number of nodes, time and time step
         common/mn/ m, n
         common/tdt/ t, dt

         m=101 
         n=11

c        Maximum time
         tmax=0.50025

         dt=0.00025


c        Set t and tout,dtout
         t=0.
c        Interval of output
         dtout=0.5  

         tout=dtout
   
c         Computation of initial values
c         for parameters and some useful
c         arrays as the mass of cells and nodes.

          call ini
       
c         Computation of parameters on the next step

1         continue

          t=t+dt

          call stepe
        
c         Output of the results  

          if(t.ge.tout) then
          call output
          tout=tout+dtout
          end if
          if(t.ge.tmax) stop

          go to 1

          end

c================================================

        subroutine ini
        implicit real*8(a-h,o-z)
c       Computation of initial data
c       and useful arrays
c        x,y - coordinates
c        wx, wy - velocities
c        rho - density
c        p - pressure
c        e - internal energy
c        vn - volume of nodes
c        vc - volume of cells
c        zmn - mass of nodes
c        zmc - mass of cells

        common/mn/ m, n

        common/xy/ x(0:201,0:201), y(0:201,0:201)
        common/rho/ rho(0:201,0:201)
        common/vareps/ e(0:201,0:201)
        common/press/ p(0:201,0:201)
        common/wxwy/ wx(0:201,0:201), wy(0:201,0:201)
        common/mass/ zmc(0:201,0:201), zmn(0:201,0:201)
        common/vol/ vc(0:201,0:201), vn(0:201,0:201)

        
c       Grid 
        hx=1./(m-1.)
        hy=0.1/(n-1.)
        pi=acos(-1.)
    
        do i=1,m
        do j=1,n
        x(i,j)=hx*(i-1.)+(n-j)*hx*sin(pi*(i-1.)/(m-1.))
        y(i,j)=hy*(j-1.)
        end do
        end do
c       
c       Fictitious nodes 
        do i=1,m
        x(i,0)=x(i,1)
        x(i,n+1)=x(i,n)
        y(i,0)=y(i,1)
        y(i,n+1)=y(i,n)
        end do
        do j=1,n
        x(0,j)=x(1,j)
        x(m+1,j)=x(m,j)
        y(0,j)=y(1,j)
        y(m+1,j)=y(m,j)
        end do
        
c       Initial profile for density and internal energy
c       and pressure
        do i=1,m-1
        do j=1,n-1
        xm=0.25*(x(i,j)+x(i+1,j)+x(i+1,j+1)+x(i,j+1))
        rho(i,j)=frho(xm)
        e(i,j)=fe(xm)
c       Pressure from the equation of state
        p(i,j)=fp(rho(i,j),e(i,j))
        end do
        end do

c       Initial profile for velocity
        do i=2,m-1
        do j=1,n
        wx(i,j)=fw(x(i,j))
        wy(i,j)=0.
        end do
        end do

c       Boundary conditions for velocity
c       Piston problem 
c       left boundary is the surface of the piston
c       and moving with constant velocity
c       wx(1,j)=1.                       
c       wy(1,j)=0.                       
c       right boundary is a stationary wall
c       wx(m,j)=0.
c       wy(m,j)=0.
c       Initial data corresponding to a 1D problem

        do j=1,n
        wx(1,j)=1.
        wx(m,j)=0.

        wy(1,j)=0.    
        wy(m,j)=0.
        end do

        do i=1,m
        wy(i,1)=0.
        wy(i,n)=0.
        end do


c        Boundary conditions for pressure
 
c        Top and bottom boundary
c        stationary walls
         do i=1,m-1
         p(i,0)=p(i,1)        
         p(i,n)=p(i,n-1)
         end do        

c        Right boundary
c        stationary wall
         do j=1,n-1
         p(m,j)=p(m-1,j)
         end do
c        Left boundary
c        Piston which is moving with CONSTANT velocity
         do j=1,n-1
         p(0,j)=p(1,j)
         end do
c


c       Volume of the cells
        do i=0,m
        do j=0,n
        vc(i,j)=0.5*((x(i+1,j)-x(i,j+1))*(y(i+1,j+1)-y(i,j))-
     *               (y(i+1,j)-y(i,j+1))*(x(i+1,j+1)-x(i,j)))
        end do
        end do


c       Mass of the cells
        do i=0,m
        do j=0,n
        zmc(i,j)=rho(i,j)*vc(i,j)
        end do
        end do

c       Mass of the nodes
        do i=1,m
        do j=1,n
        zmn(i,j)=0.25*(zmc(i-1,j-1)+zmc(i,j-1)+zmc(i,j)+zmc(i-1,j))
        end do
        end do

        return
        end 

c==================================================

         real function fp*8(rho,e)
         implicit real*8(a-h,o-z)
c       Equation of state
c       (Ideal gas gamma=5/3)

        gamma=5./3.

        fp = (gamma-1.)*rho*e

        return
        end    

c ==================================================
         real function fc*8(rho,e)
         implicit real*8(a-h,o-z)
c        Adiabatic speed of sound

         gamma=5./3.
         fc=sqrt((gamma-1.)*gamma*e)

        return
         end
c ==================================================

         real function frho*8(x)
         implicit real*8(a-h,o-z)
c       Initial profile for density
        frho = 1.
        return
        end

c ==================================================

         real function fe*8(x)
          implicit real*8(a-h,o-z)
c       Initial profile for internal energy
        fe = 0.0001      
        return
         end

c ==================================================

         real function fw*8(x)
         implicit real*8(a-h,o-z)

c        Initial profile for velocity
        fw = 0.0      

        return
         end
c ==================================================

         real function fdvdx*8(i,j,k,l)
c        derivative of volume 
c        dV(i,j)/dx(k,l)
 
c        for (k,l) = (i,j)         
c            (k,l) = (i+1,j)         
c            (k,l) = (i+1,j+1)         
c            (k,l) = (i,j+1)         

         implicit real*8(a-h,o-z)
         common/xy/ x(0:201,0:201), y(0:201,0:201)

         if(k.eq.i.and.l.eq.j) then
         fdvdx=0.5*(y(i+1,j)-y(i,j+1))
         end if
         if(k.eq.i+1.and.l.eq.j) then
         fdvdx=-0.5*(y(i,j)-y(i+1,j+1))
         end if
         if(k.eq.i+1.and.l.eq.j+1) then
         fdvdx=-0.5*(y(i+1,j)-y(i,j+1))
         end if
         if(k.eq.i.and.l.eq.j+1) then
         fdvdx=0.5*(y(i,j)-y(i+1,j+1))
         end if
     
        return
         end
c ==================================================
         real function fdvdy*8(i,j,k,l)
c        derivative of volume 
c        dV(i,j)/dy(k,l)
 
c        for (k,l) = (i,j)         
c            (k,l) = (i+1,j)         
c            (k,l) = (i+1,j+1)         
c            (k,l) = (i,j+1)         

         implicit real*8(a-h,o-z)
         common/xy/ x(0:201,0:201), y(0:201,0:201)

         if(k.eq.i.and.l.eq.j) then
         fdvdy=-0.5*(x(i+1,j)-x(i,j+1))
         end if
         if(k.eq.i+1.and.l.eq.j) then
         fdvdy=0.5*(x(i,j)-x(i+1,j+1))
         end if
         if(k.eq.i+1.and.l.eq.j+1) then
         fdvdy=0.5*(x(i+1,j)-x(i,j+1))
         end if
         if(k.eq.i.and.l.eq.j+1) then
         fdvdy=-0.5*(x(i,j)-x(i+1,j+1))
         end if
     
        return
         end
c ==========================================================
         subroutine stepe
         implicit real*8(a-h,o-z)
c       Computation of the initial data
c       and useful arrays
c        x,y - coordinates
c        wx, wy - velocities
c        rho - density
c        p - pressure
c        e - internal energy
c        vn - volume of nodes
c        vc - volume of cells
c        zmn - mass of nodes
c        zmc - mass of cells

        common/mn/ m, n
        common/tdt/ t, dt

        common/xy/ x(0:201,0:201), y(0:201,0:201)
        common/rho/ rho(0:201,0:201)
        common/vareps/ e(0:201,0:201)
        common/press/ p(0:201,0:201)
        common/wxwy/ wx(0:201,0:201), wy(0:201,0:201)
        common/mass/ zmc(0:201,0:201), zmn(0:201,0:201)
        common/vol/ vc(0:201,0:201), vn(0:201,0:201)

       

c       Velocity on n+1 - np1 time level
        dimension wxnp1(0:201,0:201)
        dimension wynp1(0:201,0:201)

       
c       Artificial viscosity
        dimension q(0:201,0:201)
        

c       Artificial viscosity
c       Linear artificial viscosity
c       Coefficient of artificial viscosity
        znu=1.  
        gamma=5./3.
        do i=1,m-1
        do j=1,n-1
c       Divergence of the velocity 
c       from the n-time level  
        DIVij=(1./vc(i,j))*
     *  (fdvdx(i,j,i,j)*wx(i,j)+fdvdx(i,j,i+1,j)*wx(i+1,j)+
     *   fdvdx(i,j,i+1,j+1)*wx(i+1,j+1)+fdvdx(i,j,i,j+1)*wx(i,j+1)+
     *   fdvdy(i,j,i,j)*wy(i,j)+fdvdy(i,j,i+1,j)*wy(i+1,j)+
     *   fdvdy(i,j,i+1,j+1)*wy(i+1,j+1)+fdvdy(i,j,i,j+1)*wy(i,j+1))
        if(DIVij.ge.0) then
        q(i,j)=0.
        else
        zfc=fc(rho(i,j), e(i,j))
        zl=sqrt(vc(i,j))
        q(i,j)=-znu*rho(i,j)* zfc*zl*DIVij
        end if
        end do
        end do
c       Boundary conditions for artificial viscosity
c       the same as for pressure
 
c        Top and bottom boundary
c        stationary walls
         do i=1,m-1
         q(i,0)=q(i,1)        
         q(i,n)=q(i,n-1)
         end do        

c        Right boundary
c        stationary wall
         do j=1,n-1
         q(m,j)=q(m-1,j)
         end do
c        Left boundary
c        Piston which is moving with CONSTANT velocity
         do j=1,m-1
         q(0,j)=q(1,j)
         end do
        
c       New velocities

c       Boundary conditions

        do j=1,n
        wxnp1(1,j)=1.
        wxnp1(m,j)=0.
        end do

        do i=1,m
        wynp1(i,1)=0.
        wynp1(i,n)=0.
        end do
     
c       internal nodes from the equation of motion
c       
        do i=2,m-1
        do j=1,n
        zpij=p(i,j)+q(i,j)
        zpim1j=p(i-1,j)+q(i-1,j)
        zpim1jm1=p(i-1,j-1)+q(i-1,j-1)
        zpijm1=p(i,j-1)+q(i,j-1)
        wxnp1(i,j) = wx(i,j)+
     *  dt*(fdvdx(i,j,i,j)*zpij+fdvdx(i-1,j,i,j)*zpim1j+
     *      fdvdx(i-1,j-1,i,j)*zpim1jm1+fdvdx(i,j-1,i,j)*zpijm1)
     *   /zmn(i,j)
        end do
        end do

        do i=1,m
        do j=2,n-1
        zpij=p(i,j)+q(i,j)
        zpim1j=p(i-1,j)+q(i-1,j)
        zpim1jm1=p(i-1,j-1)+q(i-1,j-1)
        zpijm1=p(i,j-1)+q(i,j-1)
        wynp1(i,j) = wy(i,j)+
     *  dt*(fdvdy(i,j,i,j)*zpij+fdvdy(i-1,j,i,j)*zpim1j+
     *      fdvdy(i-1,j-1,i,j)*zpim1jm1+fdvdy(i,j-1,i,j)*zpijm1)
     *   /zmn(i,j)
        end do
        end do
        
c       New coordinates

        do i=1,m
        do j=1,n
        x(i,j)=x(i,j)+0.5*dt*(wxnp1(i,j)+wx(i,j))
        y(i,j)=y(i,j)+0.5*dt*(wynp1(i,j)+wy(i,j))
        end do
        end do
c       
c       Fictitious nodes 
        do i=1,m
        x(i,0)=x(i,1)
        x(i,n+1)=x(i,n)
        y(i,0)=y(i,1)
        y(i,n+1)=y(i,n)
        end do
        do j=1,n
        x(0,j)=x(1,j)
        x(m+1,j)=x(m,j)
        y(0,j)=y(1,j)
        y(m+1,j)=y(m,j)
        end do

c       New internal energy
        do i=1,m-1
        do j=1,n-1
        wxij=0.5*(wxnp1(i,j)+wx(i,j))
        wyij=0.5*(wynp1(i,j)+wy(i,j))
        wxip1j=0.5*(wxnp1(i+1,j)+wx(i+1,j))
        wyip1j=0.5*(wynp1(i+1,j)+wy(i+1,j))
        wxip1jp1=0.5*(wxnp1(i+1,j+1)+wx(i+1,j+1))
        wyip1jp1=0.5*(wynp1(i+1,j+1)+wy(i+1,j+1))
        wxijp1=0.5*(wxnp1(i,j+1)+wx(i,j+1))
        wyijp1=0.5*(wynp1(i,j+1)+wy(i,j+1))
        e(i,j)=e(i,j)-
     *  dt*(p(i,j)+q(i,j))*
     *  (fdvdx(i,j,i,j)*wxij+fdvdx(i,j,i+1,j)*wxip1j+
     *   fdvdx(i,j,i+1,j+1)*wxip1jp1+fdvdx(i,j,i,j+1)*wxijp1+
     *   fdvdy(i,j,i,j)*wyij+fdvdy(i,j,i+1,j)*wyip1j+
     *   fdvdy(i,j,i+1,j+1)*wyip1jp1+fdvdy(i,j,i,j+1)*wyijp1)
     *  /zmc(i,j)
        end do
        end do

c       New volume 
        do i=1,m-1
        do j=1,n-1
        vc(i,j)=0.5*((x(i+1,j)-x(i,j+1))*(y(i+1,j+1)-y(i,j))-
     *               (y(i+1,j)-y(i,j+1))*(x(i+1,j+1)-x(i,j)))
        if(vc(i,j).le.0) then 
        write(*,*) 'Negative volume ', i,j
        call output
        stop
        end if
        end do
        end do

       
c        New density
         do i=1,m-1
         do j=1,n-1
         rho(i,j)=zmc(i,j)/vc(i,j)
         end do
         end do

c        New pressure
         do i=1,m-1
         do j=1,n-1
         p(i,j)=fp(rho(i,j),e(i,j))
         end do
         end do

c        Boundary conditions for pressure
 
c        Top and bottom boundary
c        stationary walls
         do i=1,m-1
         p(i,0)=p(i,1)        
         p(i,n)=p(i,n-1)
         end do        

c        Right boundary
c        stationary wall
         do j=1,n-1
         p(m,j)=p(m-1,j)
         end do
c        Left boundary
c        Piston which is moving with CONSTANT velocity
         do j=1,m-1
         p(0,j)=p(1,j)
         end do

         do i=1,m
         do j=1,n
         wx(i,j)=wxnp1(i,j)
         wy(i,j)=wynp1(i,j)
         end do
         end do


         return
         end


c================================================

         subroutine output

         implicit real*8(a-h,o-z)
c        Output of flow parameters
c        x,y - coordinates
c        wx, wy - velocities
c        rho - density
c        p - pressure
c        e - internal energy
c        vn - volume of nodes
c        vc - volume of cells
c        zmn - mass of nodes
c        zmc - mass of cells

        common/mn/ m, n
        common/tdt/ t, dt

        common/xy/ x(0:201,0:201), y(0:201,0:201)
        common/rho/ rho(0:201,0:201)
        common/vareps/ e(0:201,0:201)
        common/press/ p(0:201,0:201)
        common/wxwy/ wx(0:201,0:201), wy(0:201,0:201)
        common/mass/ zmc(0:201,0:201), zmn(0:201,0:201)


        open(unit=1,file='g2de.output')
        write(1,*) '              This is test output for program g2de'
        write(1,*)
        write(1,*) '      m=',m,' n=',n,' t=',t,' dt=',dt
        write(1,*)

        j=2
        write(1,*) '                                        j=',j
        write(1,*) 
        write(1,*)' i        x        y          wx        wy        rho
     *           e        p'
        write(1,*) 
        do i=1,m
        write(1,99) i,x(i,j),y(i,j),wx(i,j),wy(i,j),rho(i,j),
     *              e(i,j),p(i,j)
        end do
        write(1,*) 
        j=6
        write(1,*) '                                        j=',j
        write(1,*) 
        write(1,*) 
        write(1,*)' i        x        y          wx        wy        rho
     *           e        p'
        write(1,*) 
        do i=1,m
        write(1,99) i,x(i,j),y(i,j),wx(i,j),wy(i,j),rho(i,j),
     *              e(i,j),p(i,j)
        end do
        write(1,*) 
        j=10
        write(1,*) '                                        j=',j
        write(1,*) 
        write(1,*) 
        write(1,*)' i        x        y          wx        wy        rho
     *           e        p'
        write(1,*) 
        do i=1,m
        write(1,99) i,x(i,j),y(i,j),wx(i,j),wy(i,j),rho(i,j),
     *              e(i,j),p(i,j)
        end do

99      format(i3,7(1X,e10.3))

        return
        end
        


